import json
import requests


def create_deposition(access_token: str,
                      upload_type: str,
                      title: str,
                      description: str,
                      **kwargs):
    """Creates a deposition in your Zenodo account.

        Parameters
        ----------
        access_token: str
            Your Zenodo access token.
        upload_type: str

        title: str

        description: str

        kwargs:


        Returns
        -------
        deposition_id: str
            ID of the created deposition.
    """
    url = "https://zenodo.org/api/deposit/depositions"
    headers = {"Content-Type": "application/json"}

    data = {"metadata": {"upload_type": upload_type,
                         'title': title,
                         'description': description,
                         **kwargs}}

    r = requests.post(url,
                      params={'access_token': access_token},
                      data=json.dumps(data),
                      headers=headers)
    if r.status_code == 201:
        print("New Deposition has been successfully created!")
        return str(r.json()['id'])
    else:
        raise Exception(r.json()['message'])


def update_deposition(deposition_id: str,
                      access_token: str,
                      metadata: dict):
    """Updates the existing deposition with ``deposition_id`` in your Zenodo account.

        Parameters
        ----------
        deposition_id: str
            ID of a deposition in your Zenodo account.
        access_token: str
            Your Zenodo access token.
        metadata: dict

    """
    url = f"https://zenodo.org/api/deposit/depositions/{deposition_id}?access_token={access_token}"
    headers = {"Content-Type": "application/json"}

    r = requests.put(url, data=json.dumps(metadata), headers=headers)

    if r.status_code == 200:
        print("Deposition has been successfully updated!")
    else:
        raise Exception(r.json()['message'])


def delete_deposition(deposition_id: str,
                      access_token: str):
    """Deletes the existing deposition with ``deposition_id`` in your Zenodo account.

        Parameters
        ----------
        deposition_id: str
            ID of a deposition in your Zenodo account.
        access_token: str
            Your Zenodo Access token.

    """
    r = requests.delete(f'https://zenodo.org/api/deposit/depositions/{deposition_id}',
                        params={'access_token': access_token})

    if r.status_code == 201:
        print(f"Deposition with id = {deposition_id} has been successfullu deleted!")
    else:
        raise Exception(r.json()['message'])


def publish_deposition(deposition_id: str,
                       access_token: str):
    """Publishes the existing deposition with ``deposition_id`` in your Zenodo account.

        Parameters
        ----------
        deposition_id: str
            ID of a deposition in your Zenodo account.
        access_token: str
            Your Zenodo access token.

        Returns
        -------
        download_link: str
            Generated direct download link for the uploaded model in the deposition. Please **Note** that the link is usable **after** your published your deposition.
    """
    r = requests.post(f'https://zenodo.org/api/deposit/depositions/{deposition_id}/actions/publish',
                      params={'access_token': access_token})
    if r.status_code == 202:
        print(f"Deposition with id = {deposition_id} has been successfully published!")
    else:
        raise Exception(r.json()['message'])


def get_all_deposition_ids(access_token: str):
    """Gets list of all of deposition IDs existed in your Zenodo account.

        Parameters
        ----------
        access_token: str
            Your Zenodo access token.

        Returns
        -------
        deposition_ids: list
            List of deposition IDs.
    """
    r = requests.get('https://zenodo.org/api/deposit/depositions',
                     params={'access_token': access_token})

    if r.status_code != 200:
        raise Exception(r.json()['message'])

    deposition_ids = []
    for deposition_dict in r.json():
        deposition_ids.append(str(deposition_dict['id']))

    return deposition_ids




import tensorflow as tf
from tensorflow.keras import backend as K
from tensorflow.keras.layers import Activation, LeakyReLU


def mean_activation(x):
    return tf.clip_by_value(K.exp(x), 1e-5, 1e6)


def disp_activation(x):
    return tf.clip_by_value(tf.nn.softplus(x), 1e-4, 1e4)


ACTIVATIONS = {
    "relu": Activation("relu", name='reconstruction_output'),
    'leaky_relu': LeakyReLU(name="reconstruction_output"),
    'linear': Activation("linear", name='reconstruction_output'),
    'mean_activation': Activation(mean_activation, name="decoder_mean"),
    'disp_activation': Activation(disp_activation, name="decoder_disp"),
    'sigmoid': Activation('sigmoid', name='decoder_pi'),
}



import matplotlib
from matplotlib import pyplot as plt
from scarches.plotting import _alluvial


def sankey_diagram(data, save_path=None, show=False, **kwargs):
    """Draws Sankey diagram for the given ``data``.

        Parameters
        ----------
        data: :class:`~numpy.ndarray`
            array with 2 columns. One for predictions and another for true values.
        save_path: str
            Path to save the drawn Sankey diagram. if ``None``, the diagram will not be saved.
        show: bool
            if ``True`` will show the diagram.
        kwargs:
            additional arguments for diagram configuration. See ``_alluvial.plot`` function.

    """
    font = {'family': 'Arial',
            'weight': 'bold',
            'size': 14}

    matplotlib.rc('font', **font)
    matplotlib.rc('ytick', labelsize=14)
    matplotlib.rc('xtick', labelsize=14)
    plt.close('all')
    ax = _alluvial.plot(data.tolist(),
                        color_side=kwargs.get("color_side", 1),
                        alpha=kwargs.get("alpha", 0.5),
                        x_range=kwargs.get("x_range", (0, 1)),
                        res=kwargs.get("res", 20),
                        figsize=kwargs.get("figsize", (21, 15)),
                        disp_width=kwargs.get("disp_width", True),
                        width_in=kwargs.get("width_in", True),
                        wdisp_sep=kwargs.get("wdisp_sep", ' ' * 2),
                        cmap=kwargs.get("cmap", matplotlib.cm.get_cmap('jet')),
                        v_gap_frac=kwargs.get("v_gap_frac", 0.03),
                        h_gap_frac=kwargs.get("h_gap_frac", 0.03),
                        labels=kwargs.get("labels", None),
                        fontname=kwargs.get("fontname", "Arial"),
                        )
    if save_path is not None:
        plt.savefig(save_path, dpi=kwargs.get("dpi", 200), bbox_inches='tight')
    if show:
        plt.show()
    plt.close()



class Adaptor(object):
    def __init__(self, study, encoder_weights, decoder_weights):
        self.study = study

        self.encoder_weights = encoder_weights
        self.decoder_weights = decoder_weights






import tensorflow as tf
from scipy import sparse

from scarches.utils import label_encoder


def preprocess_cvae_input(n_conditions):
    def preprocess(x, y):
        x['encoder_label'] = tf.reshape(tf.one_hot(tf.cast(x['encoder_label'], tf.uint8), n_conditions), (n_conditions,))
        x['decoder_label'] = tf.reshape(tf.one_hot(tf.cast(x['decoder_label'], tf.uint8), n_conditions), (n_conditions,))

        return x, y

    return preprocess


def make_dataset(adata, condition_key, le, batch_size, n_epochs, is_training, loss_fn, n_conditions,
                 size_factor_key=None, use_mmd=False):
    if sparse.issparse(adata):
        expressions = adata.X.A
    else:
        expressions = adata.X

    encoded_conditions, le = label_encoder(adata, le, condition_key)
    if loss_fn == 'nb':
        if sparse.issparse(adata.raw.X):
            raw_expressions = adata.raw.X.A
        else:
            raw_expressions = adata.raw.X
        dataset = tf.data.Dataset.from_tensor_slices(({"expression": expressions,
                                                       "encoder_label": encoded_conditions,
                                                       "decoder_label": encoded_conditions,
                                                       "size_factor": adata.obs[size_factor_key].values},
                                                      {"reconstruction": raw_expressions}
                                                      ))
    elif loss_fn == 'zinb':
        if sparse.issparse(adata.raw.X):
            raw_expressions = adata.raw.X.A
        else:
            raw_expressions = adata.raw.X
        dataset = tf.data.Dataset.from_tensor_slices(({"expression": expressions,
                                                       "encoder_label": encoded_conditions,
                                                       "decoder_label": encoded_conditions,
                                                       'size_factor': adata.obs[size_factor_key].values},
                                                      {"reconstruction": raw_expressions}
                                                      ))
    else:
        if use_mmd:
            dataset = tf.data.Dataset.from_tensor_slices(({"expression": expressions,
                                                           "encoder_label": encoded_conditions,
                                                           "decoder_label": encoded_conditions},
                                                          {"reconstruction": expressions,
                                                           "mmd": encoded_conditions}
                                                          ))
        else:
            dataset = tf.data.Dataset.from_tensor_slices(({"expression": expressions,
                                                           "encoder_label": encoded_conditions,
                                                           "decoder_label": encoded_conditions},
                                                          {"reconstruction": expressions}
                                                          ))
    if is_training:
        dataset = dataset.shuffle(1000)
    dataset = dataset.map(preprocess_cvae_input(n_conditions),
                          num_parallel_calls=4,
                          deterministic=None)
    if is_training:
        dataset = dataset.repeat(n_epochs)
    else:
        dataset = dataset.repeat()

    dataset = dataset.batch(batch_size, drop_remainder=True if is_training else False)
    dataset = dataset.prefetch(buffer_size=5)

    return dataset, le



import os

import anndata
import numpy as np
import tensorflow as tf
from tensorflow import keras
from tensorflow.keras.callbacks import EarlyStopping, History, ReduceLROnPlateau, LambdaCallback
from tensorflow.keras.layers import Dense, BatchNormalization, Dropout, Lambda, LeakyReLU
from tensorflow.keras.models import Model
from tensorflow.keras.utils import to_categorical
from tensorflow.keras.utils import get_custom_objects
from scipy import sparse

from scarches.models import CVAE
from scarches.models._activations import ACTIVATIONS
from scarches.models._callbacks import ScoreCallback
from scarches.models._data_generator import make_dataset
from scarches.models._layers import LAYERS
from scarches.models._losses import LOSSES
from scarches.models._utils import print_progress
from scarches.utils import label_encoder, remove_sparsity, train_test_split


class scArches(CVAE):
    """scArches class. This class contains the implementation of scArches network.

        Parameters
        ----------
        x_dimension: int
            number of gene expression space dimensions.
        conditions: list
            list of unique conditions (i.e. batch ids) in the data used for one-hot encoding.
        z_dimension: int
            number of latent space dimensions.
        task_name: str
            name of the task.

        kwargs:
            `learning_rate`: float
                scArches's optimizer's step size (learning rate).
            `alpha`: float
                KL divergence coefficient in the loss function.
            `beta`: float
                MMD loss coefficient in the loss function.
            `eta`: float
                Reconstruction coefficient in the loss function.
            `dropout_rate`: float
                dropout rate for Dropout layers in scArches' architecture.
            `model_path`: str
                path to save model config and its weights.
            `clip_value`: float
                Optimizer's clip value used for clipping the computed gradients.
            `output_activation`: str
                Output activation of scArches which Depends on the range of data. For positive value you can use "relu" or "linear" but if your data
                have negative value set to "linear" which is default.
            `use_batchnorm`: bool
                Whether use batch normalization in scArches or not.
            `architecture`: list
                Architecture of scArches. Must be a list of integers.
            `gene_names`: list
                names of genes fed as scArches' input. Must be a list of strings.
    """

    def __new__(cls, *args, **kwargs):
        loss_fn = kwargs.get("loss_fn", "nb")
        if loss_fn in ['nb', 'zinb']:
            if loss_fn == 'nb':
                from .scarchesnb import scArchesNB
                return scArchesNB(*args, **kwargs)
            elif loss_fn == 'zinb':
                from .scarcheszinb import scArchesZINB
                return scArchesZINB(*args, **kwargs)
        elif kwargs.get('beta', 0.0) == 0:
            from .cvae import CVAE
            if kwargs.get("beta", None) is not None:
                kwargs.pop("beta")
            return CVAE(*args, **kwargs)
        else:
            return super(scArches, cls).__new__(cls)

    def __init__(self, x_dimension, conditions, task_name="unknown", z_dimension=10, **kwargs):
        self.beta = kwargs.pop('beta', 0.0)
        self.mmd_computation_method = kwargs.pop("mmd_computation_method", "general")

        if kwargs.get("loss_fn", "mse") in ['nb', 'zinb']:
            kwargs['loss_fn'] = 'mse'

        kwargs.update({"model_name": "cvae", "class_name": "scArches"})

        super().__init__(x_dimension, conditions, task_name, z_dimension, **kwargs)

        self.network_kwargs.update({
            "mmd_computation_method": self.mmd_computation_method,
        })

        self.training_kwargs.update({
            "beta": self.beta,
        })

    def update_kwargs(self):
        super().update_kwargs()
        self.network_kwargs.update({
            "mmd_computation_method": self.mmd_computation_method,
        })

        self.training_kwargs.update({
            "beta": self.beta,
        })

    @classmethod
    def from_config(cls, config_path, new_params=None, compile=True, construct=True):
        """create class object from exsiting class' config file.

        Parameters
        ----------
        config_path: str
            Path to scArches' config json file.
        new_params: dict, optional
            Python dict of parameters which you wanted to assign new values to them.
        compile: bool
            ``True`` by default. if ``True``, will compile scArches' model after creating an instance.
        construct: bool
            ``True`` by default. if ``True``, will construct scArches' model after creating an instance.
        """
        import json
        with open(config_path, 'rb') as f:
            class_config = json.load(f)

        class_config['construct_model'] = construct
        class_config['compile_model'] = compile

        if new_params:
            class_config.update(new_params)

        return cls(**class_config)

    def _decoder(self, name="decoder"):
        for idx, n_neuron in enumerate(self.architecture[::-1]):
            if idx == 0:
                h = LAYERS['FirstLayer'](n_neuron, kernel_initializer=self.init_w,
                                         use_bias=False, name="first_layer", freeze=self.freeze_expression_input)(
                    [self.z, self.decoder_labels])
            else:
                h = Dense(n_neuron, kernel_initializer=self.init_w,
                          use_bias=False)(h)
            if self.use_batchnorm:
                h = BatchNormalization()(h)
            h = LeakyReLU()(h)
            if idx == 0:
                h_mmd = h
            h = Dropout(self.dr_rate)(h)
        model_inputs, model_outputs = self._output_decoder(h)
        self.decoder = Model(inputs=model_inputs, outputs=model_outputs, name=name)
        self.mmd_decoder = Model(inputs=model_inputs, outputs=h_mmd, name='mmd_decoder')

    def call(self, x, training=None, mask=None):
        if isinstance(x, list):
            expression, encoder_labels, decoder_labels = x
        else:
            expression = x['expression']
            encoder_labels = x['encoder_label']
            decoder_labels = x['decoder_label']

        z_mean, z_log_var, z = self.encoder([expression, encoder_labels])

        x_hat = self.decoder([z, decoder_labels])
        mmd_output = self.mmd_decoder([z, decoder_labels])
        return x_hat, mmd_output, z_mean, z_log_var

    def calc_losses(self, y_true, y_pred, mmd_true, mmd_pred, z_mean, z_log_var):
        """
            Defines the loss function of class' network after constructing the whole
            network.
        """
        recon_loss = LOSSES[f'{self.loss_fn}_recon'](y_true, y_pred)
        mmd_loss = LOSSES['mmd'](self.n_conditions)(mmd_true, mmd_pred)
        kl_loss = LOSSES['kl'](z_mean, z_log_var)(y_true, y_pred)

        loss = self.eta * recon_loss + self.alpha * kl_loss + self.beta * mmd_loss

        return loss, self.eta * recon_loss, kl_loss, self.beta * mmd_loss

    def train_step(self, data):
        with tf.GradientTape() as tape:
            loss, recon_loss, kl_loss, mmd_loss = self.forward_with_loss(data)

        grads = tape.gradient(loss, self.trainable_weights)
        self.optimizer.apply_gradients(zip(grads, self.trainable_weights))
        return {
            "loss": loss,
            f'{self.loss_fn}_loss': recon_loss,
            "kl_loss": kl_loss,
            "mmd_loss": mmd_loss
        }

    def forward_with_loss(self, data):
        x, y = data

        y_true, mmd_true = y['reconstruction'], y['mmd']

        y_pred, mmd_output, z_mean, z_log_var = self.call(x)
        loss, recon_loss, kl_loss, mmd_loss = self.calc_losses(y_true, y_pred, mmd_true, mmd_output, z_mean, z_log_var)

        return loss, recon_loss, kl_loss, mmd_loss

    def test_step(self, data):
        loss, recon_loss, kl_loss, mmd_loss = self.forward_with_loss(data)

        return {
            'loss': loss,
            f'{self.loss_fn}_loss': recon_loss,
            'kl_loss': kl_loss,
            'mmd_loss': mmd_loss,
        }

    def to_mmd_layer(self, adata, batch_key):
        """
            Map ``adata`` in to the MMD space. This function will feed data
            in ``mmd_model`` of scArches and compute the MMD space coordinates
            for each sample in data.

            Parameters
            ----------
            adata: :class:`~anndata.AnnData`
                Annotated data matrix to be mapped to MMD latent space.
                Please note that ``adata.X`` has to be in shape [n_obs, x_dimension]
            encoder_labels: :class:`~numpy.ndarray`
                :class:`~numpy.ndarray` of labels to be fed as scArches' encoder condition array.
            decoder_labels: :class:`~numpy.ndarray`
                :class:`~numpy.ndarray` of labels to be fed as scArches' decoder condition array.

            Returns
            -------
            adata_mmd: :class:`~anndata.AnnData`
                returns Annotated data containing MMD latent space encoding of ``adata``
        """
        adata = remove_sparsity(adata)

        encoder_labels, _ = label_encoder(adata, self.condition_encoder, batch_key)
        decoder_labels, _ = label_encoder(adata, self.condition_encoder, batch_key)

        encoder_labels = to_categorical(encoder_labels, num_classes=self.n_conditions)
        decoder_labels = to_categorical(decoder_labels, num_classes=self.n_conditions)

        cvae_inputs = [adata.X, encoder_labels, decoder_labels]

        mmd = self(cvae_inputs)[1].numpy()
        mmd = np.nan_to_num(mmd, nan=0.0, posinf=0.0, neginf=0.0)

        adata_mmd = anndata.AnnData(X=mmd)
        adata_mmd.obs = adata.obs.copy(deep=True)

        return adata_mmd

    def get_latent(self, adata, batch_key, return_z=True):
        """ Transforms `adata` in latent space of scArches and returns the latent
        coordinates in the annotated (adata) format.

        Parameters
        ----------
        adata: :class:`~anndata.AnnData`
            Annotated dataset matrix in Primary space.
        batch_key: str
            Name of the column containing the study (batch) names for each sample.
        return_z: bool
            ``False`` by defaul. if ``True``, the output of bottleneck layer of network will be computed.

        Returns
        -------
        adata_pred: `~anndata.AnnData`
            Annotated data of transformed ``adata`` into latent space.
        """
        if set(self.gene_names).issubset(set(adata.var_names)):
            adata = adata[:, self.gene_names]
        else:
            raise Exception("set of gene names in train adata are inconsistent with scArches' gene_names")

        if self.beta == 0:
            return_z = True

        encoder_labels, _ = label_encoder(adata, self.condition_encoder, batch_key)
        encoder_labels = to_categorical(encoder_labels, num_classes=self.n_conditions)

        if return_z or self.beta == 0:
            return self.get_z_latent(adata, encoder_labels)
        else:
            return self.to_mmd_layer(adata, batch_key)

    def predict(self, adata, encoder_labels, decoder_labels):
        """Feeds ``adata`` to scArches and produces the reconstructed data.

            Parameters
            ----------
            adata: :class:`~anndata.AnnData`
                Annotated data matrix whether in primary space.
            encoder_labels: :class:`~numpy.ndarray`
                :class:`~numpy.ndarray` of labels to be fed as scArches' encoder condition array.
            decoder_labels: :class:`~numpy.ndarray`
                :class:`~numpy.ndarray` of labels to be fed as scArches' decoder condition array.

            Returns
            -------
            adata_pred: `~anndata.AnnData`
                Annotated data of predicted cells in primary space.
        """
        adata = remove_sparsity(adata)

        encoder_labels = to_categorical(encoder_labels, num_classes=self.n_conditions)
        decoder_labels = to_categorical(decoder_labels, num_classes=self.n_conditions)

        x_hat = self.cvae_model.predict([adata.X, encoder_labels, decoder_labels])[0]

        adata_pred = anndata.AnnData(X=x_hat)
        adata_pred.obs = adata.obs
        adata_pred.var_names = adata.var_names

        return adata_pred

    def _fit_dataset(self, adata,
                     condition_key, train_size=0.8, cell_type_key='cell_type',
                     n_epochs=300, batch_size=512, steps_per_epoch=100,
                     early_stop_limit=10, lr_reducer=7,
                     n_per_epoch=0, score_filename=None,
                     save=True, retrain=True, verbose=3):
        train_adata, valid_adata = train_test_split(adata, train_size)

        if self.gene_names is None:
            self.gene_names = train_adata.var_names.tolist()
        else:
            if set(self.gene_names).issubset(set(train_adata.var_names)):
                train_adata = train_adata[:, self.gene_names]
            else:
                raise Exception("set of gene names in train adata are inconsistent with class' gene_names")

            if set(self.gene_names).issubset(set(valid_adata.var_names)):
                valid_adata = valid_adata[:, self.gene_names]
            else:
                raise Exception("set of gene names in valid adata are inconsistent with class' gene_names")

        if not retrain and os.path.exists(os.path.join(self.model_path, f"{self.model_name}.h5")):
            self.restore_model_weights()
            return

        callbacks = [
            History(),
        ]

        if verbose > 2:
            callbacks.append(
                LambdaCallback(on_epoch_end=lambda epoch, logs: print_progress(epoch, logs, n_epochs)))
            fit_verbose = 0
        else:
            fit_verbose = verbose

        if (n_per_epoch > 0 or n_per_epoch == -1) and not score_filename:
            adata = train_adata.concatenate(valid_adata)

            train_celltypes_encoded, _ = label_encoder(train_adata, le=None, condition_key=cell_type_key)
            valid_celltypes_encoded, _ = label_encoder(valid_adata, le=None, condition_key=cell_type_key)
            celltype_labels = np.concatenate([train_celltypes_encoded, valid_celltypes_encoded], axis=0)

            callbacks.append(ScoreCallback(score_filename, adata, condition_key, cell_type_key, self.cvae_model,
                                           n_per_epoch=n_per_epoch, n_batch_labels=len(self.n_conditions),
                                           n_celltype_labels=len(np.unique(celltype_labels))))

        if early_stop_limit > 0:
            callbacks.append(EarlyStopping(patience=early_stop_limit, monitor='val_loss'))

        if lr_reducer > 0:
            callbacks.append(ReduceLROnPlateau(monitor='val_loss', patience=lr_reducer))

        train_dataset, self.condition_encoder = make_dataset(train_adata, condition_key, self.condition_encoder,
                                                             batch_size, n_epochs, True,
                                                             self.loss_fn, self.n_conditions, use_mmd=True)
        valid_dataset, _ = make_dataset(valid_adata, condition_key, self.condition_encoder, valid_adata.shape[0],
                                        n_epochs, False,
                                        self.loss_fn, self.n_conditions, use_mmd=True)

        self.fit(train_dataset,
                 validation_data=valid_dataset,
                 epochs=n_epochs,
                 batch_size=batch_size,
                 verbose=fit_verbose,
                 callbacks=callbacks,
                 steps_per_epoch=steps_per_epoch,
                 validation_steps=1,
                 )
        if save:
            self.update_kwargs()
            self.save(make_dir=True)

    def _fit(self, adata,
             condition_key, train_size=0.8, cell_type_key='cell_type',
             n_epochs=300, batch_size=512,
             early_stop_limit=10, lr_reducer=7,
             n_per_epoch=0, score_filename=None,
             save=True, retrain=True, verbose=3):
        train_adata, valid_adata = train_test_split(adata, train_size)

        if self.gene_names is None:
            self.gene_names = train_adata.var_names.tolist()
        else:
            if set(self.gene_names).issubset(set(train_adata.var_names)):
                train_adata = train_adata[:, self.gene_names]
            else:
                raise Exception("set of gene names in train adata are inconsistent with class' gene_names")

            if set(self.gene_names).issubset(set(valid_adata.var_names)):
                valid_adata = valid_adata[:, self.gene_names]
            else:
                raise Exception("set of gene names in valid adata are inconsistent with class' gene_names")

        train_expr = train_adata.X.A if sparse.issparse(train_adata.X) else train_adata.X
        valid_expr = valid_adata.X.A if sparse.issparse(valid_adata.X) else valid_adata.X

        train_conditions_encoded, self.condition_encoder = label_encoder(train_adata, le=self.condition_encoder,
                                                                         condition_key=condition_key)

        valid_conditions_encoded, self.condition_encoder = label_encoder(valid_adata, le=self.condition_encoder,
                                                                         condition_key=condition_key)

        if not retrain and os.path.exists(os.path.join(self.model_path, f"{self.model_name}.h5")):
            self.restore_model_weights()
            return

        callbacks = [
            History(),
        ]

        if verbose > 2:
            callbacks.append(
                LambdaCallback(on_epoch_end=lambda epoch, logs: print_progress(epoch, logs, n_epochs)))
            fit_verbose = 0
        else:
            fit_verbose = verbose

        if (n_per_epoch > 0 or n_per_epoch == -1) and not score_filename:
            adata = train_adata.concatenate(valid_adata)

            train_celltypes_encoded, _ = label_encoder(train_adata, le=None, condition_key=cell_type_key)
            valid_celltypes_encoded, _ = label_encoder(valid_adata, le=None, condition_key=cell_type_key)
            celltype_labels = np.concatenate([train_celltypes_encoded, valid_celltypes_encoded], axis=0)

            callbacks.append(ScoreCallback(score_filename, adata, condition_key, cell_type_key, self.cvae_model,
                                           n_per_epoch=n_per_epoch, n_batch_labels=len(self.n_conditions),
                                           n_celltype_labels=len(np.unique(celltype_labels))))

        if early_stop_limit > 0:
            callbacks.append(EarlyStopping(patience=early_stop_limit, monitor='val_loss'))

        if lr_reducer > 0:
            callbacks.append(ReduceLROnPlateau(monitor='val_loss', patience=lr_reducer))

        train_conditions_onehot = to_categorical(train_conditions_encoded, num_classes=self.n_conditions)
        valid_conditions_onehot = to_categorical(valid_conditions_encoded, num_classes=self.n_conditions)

        x_train = [train_expr, train_conditions_onehot, train_conditions_onehot]
        x_valid = [valid_expr, valid_conditions_onehot, valid_conditions_onehot]

        y_train = [train_expr, train_conditions_encoded]
        y_valid = [valid_expr, valid_conditions_encoded]

        self.fit(x=x_train,
                 y=y_train,
                 validation_data=(x_valid, y_valid),
                 epochs=n_epochs,
                 batch_size=batch_size,
                 verbose=fit_verbose,
                 callbacks=callbacks,
                 )
        if save:
            self.update_kwargs()
            self.save(make_dir=True)

    # def _train_on_batch(self, adata,
    #                     condition_key, train_size=0.8, cell_type_key='cell_type',
    #                     n_epochs=300, batch_size=512,
    #                     early_stop_limit=10, lr_reducer=7,
    #                     n_per_epoch=0, score_filename=None,
    #                     n_gradients_per_epoch=100,
    #                     save=True, retrain=True, verbose=3):
    #     print("TRAIN_ON_BATCH")
    #     train_adata, valid_adata = train_test_split(adata, train_size)
    #
    #     if self.gene_names is None:
    #         self.gene_names = train_adata.var_names.tolist()
    #     else:
    #         if set(self.gene_names).issubset(set(train_adata.var_names)):
    #             train_adata = train_adata[:, self.gene_names]
    #         else:
    #             raise Exception("set of gene names in train adata are inconsistent with class' gene_names")
    #
    #         if set(self.gene_names).issubset(set(valid_adata.var_names)):
    #             valid_adata = valid_adata[:, self.gene_names]
    #         else:
    #             raise Exception("set of gene names in valid adata are inconsistent with class' gene_names")
    #
    #     train_conditions_encoded, self.condition_encoder = label_encoder(train_adata, le=self.condition_encoder,
    #                                                                      condition_key=condition_key)
    #
    #     valid_conditions_encoded, self.condition_encoder = label_encoder(valid_adata, le=self.condition_encoder,
    #                                                                      condition_key=condition_key)
    #
    #     if not retrain and os.path.exists(os.path.join(self.model_path, f"{self.model_name}.h5")):
    #         self.restore_model_weights()
    #         return
    #
    #     train_conditions_onehot = to_categorical(train_conditions_encoded, num_classes=self.n_conditions)
    #     valid_conditions_onehot = to_categorical(valid_conditions_encoded, num_classes=self.n_conditions)
    #
    #     if sparse.issparse(train_adata.X):
    #         is_sparse = True
    #     else:
    #         is_sparse = False
    #
    #     train_expr = train_adata.X
    #     valid_expr = valid_adata.X.A if is_sparse else valid_adata.X
    #     x_valid = [valid_expr, valid_conditions_onehot, valid_conditions_onehot]
    #
    #     if self.loss_fn in ['nb', 'zinb']:
    #         x_valid.append(valid_adata.obs[self.size_factor_key].values)
    #         y_valid = [valid_adata.raw.X.A if sparse.issparse(valid_adata.raw.X) else valid_adata.raw.X,
    #                    valid_conditions_encoded]
    #     else:
    #         y_valid = [valid_expr, valid_conditions_encoded]
    #
    #     es_patience, best_val_loss = 0, 1e10
    #     for i in range(n_epochs):
    #         train_loss = train_recon_loss = train_mmd_loss = 0.0
    #         for j in range(min(n_gradients_per_epoch, train_adata.shape[0] // batch_size)):
    #             batch_indices = np.random.choice(train_adata.shape[0], batch_size)
    #
    #             batch_expr = train_expr[batch_indices, :].A if is_sparse else train_expr[batch_indices, :]
    #
    #             x_train = [batch_expr, train_conditions_onehot[batch_indices], train_conditions_onehot[batch_indices]]
    #
    #             if self.loss_fn in ['nb', 'zinb']:
    #                 x_train.append(train_adata.obs[self.size_factor_key].values[batch_indices])
    #                 y_train = [train_adata.raw.X[batch_indices].A if sparse.issparse(
    #                     train_adata.raw.X[batch_indices]) else train_adata.raw.X[batch_indices],
    #                            train_conditions_encoded[batch_indices]]
    #             else:
    #                 y_train = [batch_expr, train_conditions_encoded[batch_indices]]
    #
    #             batch_loss, batch_recon_loss, batch_kl_loss = self.cvae_model.train_on_batch(x_train, y_train)
    #
    #             train_loss += batch_loss / batch_size
    #             train_recon_loss += batch_recon_loss / batch_size
    #             train_mmd_loss += batch_kl_loss / batch_size
    #
    #         valid_loss, valid_recon_loss, valid_mmd_loss = self.cvae_model.evaluate(x_valid, y_valid, verbose=0)
    #
    #         if valid_loss < best_val_loss:
    #             best_val_loss = valid_loss
    #             es_patience = 0
    #         else:
    #             es_patience += 1
    #             if es_patience == early_stop_limit:
    #                 print("Training stopped with Early Stopping")
    #                 break
    #
    #         logs = {"loss": train_loss, "recon_loss": train_recon_loss, "mmd_loss": train_mmd_loss,
    #                 "val_loss": valid_loss, "val_recon_loss": valid_recon_loss, "val_mmd_loss": valid_mmd_loss}
    #         print_progress(i, logs, n_epochs)
    #
    #     if save:
    #         self.update_kwargs()
    #         self.save(make_dir=True)



from .cvae import CVAE
from .scarches import scArches
from .scarchesnb import scArchesNB
from .scarcheszinb import scArchesZINB
from ._adaptor import Adaptor
__all__ = (
    'scArches', 'CVAE', 'scArchesNB', 'scArchesZINB'
)



import scarches as sca

condition_key = 'study'
adata = sca.datasets.pancreas()
print(adata.X.min(), adata.X.max())

condition_labels = adata.obs[condition_key].unique().tolist()
network = sca.models.scArches(task_name='test',
                              x_dimension=adata.shape[1],
                              z_dimension=10,
                              device='gpu',
                              loss_fn='nb',
                              architecture=[128, 32],
                              conditions=condition_labels,
                              model_path='./models/')

network.train(adata,
              n_epochs=150,
              condition_key=condition_key,
              steps_per_epoch=100,
              batch_size=128)
# network.save()
#
# new_network = sca.models.scArches.from_config('./models/test/scArchesNB.json')
# new_network.restore_model_config()
# print(new_network.model_path)



import os
import numpy as np

from scarches.models import CVAE
from tensorflow.keras.layers import Dense, Input
from tensorflow.keras.utils import get_custom_objects

import tensorflow as tf

from scarches.models._activations import ACTIVATIONS
from scarches.models._callbacks import ScoreCallback
from scarches.models._data_generator import make_dataset
from scarches.models._layers import LAYERS
from scarches.models._losses import LOSSES

from scipy import sparse
from tensorflow.keras.callbacks import History, EarlyStopping, ReduceLROnPlateau, LambdaCallback
from tensorflow.keras.utils import to_categorical

from scarches.models._utils import print_progress
from scarches.utils import train_test_split, label_encoder


class scArchesZINB(CVAE):
    """
        scArches network with ZINB for loss function class. This class contains the implementation of scNet network.

        Parameters
        ----------
        x_dimension: int
            number of gene expression space dimensions.
        n_conditions: list
            list of unique conditions (i.e. batch ids) in the data used for one-hot encoding.
        z_dimension: int
            number of latent space dimensions.
        task_name: str
            name of the task.

        kwargs:
            `learning_rate`: float
                scNet's optimizer's step size (learning rate).
            `alpha`: float
                KL divergence coefficient in the loss function.
            `eta`: float
                Reconstruction coefficient in the loss function.
            `dropout_rate`: float
                dropout rate for Dropout layers in scNet's architecture.
            `model_path`: str
                path to save model config and its weights.
            `clip_value`: float
                Optimizer's clip value used for clipping the computed gradients.
            `output_activation`: str
                Output activation of scNet which Depends on the range of data.
            `use_batchnorm`: bool
                Whether use batch normalization in scNet or not.
            `architecture`: list
                Architecture of scNet. Must be a list of integers.
            `gene_names`: list
                names of genes fed as scNet's input. Must be a list of strings.
    """

    def __init__(self, x_dimension, conditions, task_name="unknown", z_dimension=10, **kwargs):
        kwargs.update({'loss_fn': 'zinb',
                       "model_name": "cvae_zinb", "class_name": "scArchesZINB"})

        self.size_factor_key = kwargs.pop("size_factor_key", 'size_factors')
        self.ridge = kwargs.pop("ridge", 0.1)
        self.size_factor = Input(shape=(1,), name='size_factor')

        super().__init__(x_dimension, conditions, task_name, z_dimension, **kwargs)

        self.network_kwargs.update({
            "size_factor_key": self.size_factor_key,
        })

        self.training_kwargs.update({
            "ridge": self.ridge,
        })

    def update_kwargs(self):
        super().update_kwargs()
        self.network_kwargs.update({
            "size_factor_key": self.size_factor_key,
        })

        self.training_kwargs.update({
            "ridge": self.ridge,
        })

    def _output_decoder(self, h):
        h_pi = Dense(self.x_dim, activation=ACTIVATIONS['sigmoid'], kernel_initializer=self.init_w, use_bias=True,
                     name='decoder_pi')(h)
        h_mean = Dense(self.x_dim, activation=None, kernel_initializer=self.init_w,
                       use_bias=True)(h)
        h_mean = ACTIVATIONS['mean_activation'](h_mean)

        h_disp = Dense(self.x_dim, activation=None, kernel_initializer=self.init_w,
                       use_bias=True)(h)
        h_disp = ACTIVATIONS['disp_activation'](h_disp)

        mean_output = LAYERS['ColWiseMultLayer']()([h_mean, self.size_factor])

        model_outputs = [mean_output, h_disp, h_pi]

        model_inputs = [self.z, self.decoder_labels, self.size_factor]

        return model_inputs, model_outputs

    @classmethod
    def from_config(cls, config_path, new_params=None, compile=True, construct=True):
        """create CVAE_ZINB object from exsiting CVAE_ZINB's config file.

        Parameters
        ----------
        config_path: str
            Path to class' config json file.
        new_params: dict, optional
            Python dict of parameters which you wanted to assign new values to them.
        compile: bool
            ``True`` by default. if ``True``, will compile class' model after creating an instance.
        construct: bool
            ``True`` by default. if ``True``, will construct class' model after creating an instance.
        """
        import json
        with open(config_path, 'rb') as f:
            class_config = json.load(f)

        class_config['construct_model'] = construct
        class_config['compile_model'] = compile

        if new_params:
            class_config.update(new_params)

        return cls(**class_config)

    def call(self, x, training=None, mask=None):
        if isinstance(x, list):
            expression, encoder_labels, decoder_labels, size_factors = x
        else:
            expression = x['expression']
            encoder_labels = x['encoder_label']
            decoder_labels = x['decoder_label']
            size_factors = x['size_factor']

        z_mean, z_log_var, z = self.encoder([expression, encoder_labels])

        x_hat, disp, pi = self.decoder([z, decoder_labels, size_factors])
        return x_hat, z_mean, z_log_var, disp, pi

    def construct_network(self):
        """
            Constructs the whole class' network. It is step-by-step constructing the scNet
            network. First, It will construct the encoder part and get mu, log_var of
            latent space. Second, It will sample from the latent space to feed the
            decoder part in next step. Finally, It will reconstruct the data by
            constructing decoder part of scNet.
        """
        self._encoder("encoder")
        self._decoder("decoder")

        self.custom_objects = {'mean_activation': ACTIVATIONS['mean_activation'],
                               'disp_activation': ACTIVATIONS['disp_activation'],
                               'ColwiseMultLayer': LAYERS['ColWiseMultLayer'],
                               'FirstLayer': LAYERS['FirstLayer']}

        # Building the model via calling it with a random input
        input_arr = [tf.random.uniform((1, self.x_dim)), tf.ones((1, self.n_conditions)),
                     tf.ones((1, self.n_conditions)), tf.ones(1, dtype=tf.float32)]
        self(input_arr)

        get_custom_objects().update(self.custom_objects)
        print(f"{self.class_name}'s network has been successfully constructed!")

    def calc_losses(self, y_true, y_pred, z_mean, z_log_var, disp=None, pi=None):
        """
            Defines the loss function of class' network after constructing the whole
            network.
        """
        loss = LOSSES[self.loss_fn](pi, disp, z_mean, z_log_var, self.ridge,
                                    self.alpha, self.eta)(y_true, y_pred)
        recon_loss = LOSSES[f'{self.loss_fn}_wo_kl'](pi, disp, self.ridge, self.eta)(y_true, y_pred)
        kl_loss = LOSSES['kl'](z_mean, z_log_var)(y_true, y_pred)

        return loss, recon_loss, kl_loss

    def forward_with_loss(self, data):
        x, y = data
    
        y = y['reconstruction']    
        y_pred, z_mean, z_log_var, disp, pi = self.call(x)
        loss, recon_loss, kl_loss = self.calc_losses(y, y_pred, z_mean, z_log_var, disp, pi)

        return loss, recon_loss, kl_loss

    def _fit(self, adata,
             condition_key, train_size=0.8, cell_type_key='cell_type',
             n_epochs=100, batch_size=128,
             early_stop_limit=10, lr_reducer=8,
             n_per_epoch=0, score_filename=None,
             save=True, retrain=True, verbose=3):
        train_adata, valid_adata = train_test_split(adata, train_size)

        if self.gene_names is None:
            self.gene_names = train_adata.var_names.tolist()
        else:
            if set(self.gene_names).issubset(set(train_adata.var_names)):
                train_adata = train_adata[:, self.gene_names]
            else:
                raise Exception("set of gene names in train adata are inconsistent with class' gene_names")

            if set(self.gene_names).issubset(set(valid_adata.var_names)):
                valid_adata = valid_adata[:, self.gene_names]
            else:
                raise Exception("set of gene names in valid adata are inconsistent with class' gene_names")

        train_raw_expr = train_adata.raw.X.A if sparse.issparse(train_adata.raw.X) else train_adata.raw.X
        valid_raw_expr = valid_adata.raw.X.A if sparse.issparse(valid_adata.raw.X) else valid_adata.raw.X

        train_expr = train_adata.X.A if sparse.issparse(train_adata.X) else train_adata.X
        valid_expr = valid_adata.X.A if sparse.issparse(valid_adata.X) else valid_adata.X

        train_conditions_encoded, self.condition_encoder = label_encoder(train_adata, le=self.condition_encoder,
                                                                         condition_key=condition_key)

        valid_conditions_encoded, self.condition_encoder = label_encoder(valid_adata, le=self.condition_encoder,
                                                                         condition_key=condition_key)

        if not retrain and os.path.exists(os.path.join(self.model_path, f"{self.model_name}.h5")):
            self.restore_model_weights()
            return

        callbacks = [
            History(),
        ]

        if verbose > 2:
            callbacks.append(
                LambdaCallback(on_epoch_end=lambda epoch, logs: print_progress(epoch, logs, n_epochs)))
            fit_verbose = 0
        else:
            fit_verbose = verbose

        if (n_per_epoch > 0 or n_per_epoch == -1) and not score_filename:
            adata = train_adata.concatenate(valid_adata)

            train_celltypes_encoded, _ = label_encoder(train_adata, le=None, condition_key=cell_type_key)
            valid_celltypes_encoded, _ = label_encoder(valid_adata, le=None, condition_key=cell_type_key)
            celltype_labels = np.concatenate([train_celltypes_encoded, valid_celltypes_encoded], axis=0)

            callbacks.append(ScoreCallback(score_filename, adata, condition_key, cell_type_key, self.cvae,
                                           n_per_epoch=n_per_epoch, n_batch_labels=len(self.n_conditions),
                                           n_celltype_labels=len(np.unique(celltype_labels))))

        if early_stop_limit > 0:
            callbacks.append(EarlyStopping(patience=early_stop_limit, monitor='val_loss'))

        if lr_reducer > 0:
            callbacks.append(ReduceLROnPlateau(monitor='val_loss', patience=lr_reducer))

        train_conditions_onehot = to_categorical(train_conditions_encoded, num_classes=self.n_conditions)
        valid_conditions_onehot = to_categorical(valid_conditions_encoded, num_classes=self.n_conditions)

        x_train = [train_expr, train_conditions_onehot, train_conditions_onehot,
                   train_adata.obs[self.size_factor_key].values]
        x_valid = [valid_expr, valid_conditions_onehot, valid_conditions_onehot,
                   valid_adata.obs[self.size_factor_key].values]

        y_train = train_raw_expr
        y_valid = valid_raw_expr

        self.fit(x=x_train,
                 y=y_train,
                 validation_data=(x_valid, y_valid),
                 epochs=n_epochs,
                 batch_size=batch_size,
                 verbose=fit_verbose,
                 callbacks=callbacks,
                 )
        if save:
            self.update_kwargs()
            self.save(make_dir=True)

    def _fit_dataset(self, adata,
                     condition_key, train_size=0.8, cell_type_key='cell_type',
                     n_epochs=100, batch_size=128, steps_per_epoch=100,
                     early_stop_limit=10, lr_reducer=8,
                     n_per_epoch=0, score_filename=None,
                     save=True, retrain=True, verbose=3):
        train_adata, valid_adata = train_test_split(adata, train_size)

        if self.gene_names is None:
            self.gene_names = train_adata.var_names.tolist()
        else:
            if set(self.gene_names).issubset(set(train_adata.var_names)):
                train_adata = train_adata[:, self.gene_names]
            else:
                raise Exception("set of gene names in train adata are inconsistent with class' gene_names")

            if set(self.gene_names).issubset(set(valid_adata.var_names)):
                valid_adata = valid_adata[:, self.gene_names]
            else:
                raise Exception("set of gene names in valid adata are inconsistent with class' gene_names")

        if not retrain and os.path.exists(os.path.join(self.model_path, f"{self.model_name}.h5")):
            self.restore_model_weights()
            return

        callbacks = [
            History(),
        ]

        if verbose > 2:
            callbacks.append(
                LambdaCallback(on_epoch_end=lambda epoch, logs: print_progress(epoch, logs, n_epochs)))
            fit_verbose = 0
        else:
            fit_verbose = verbose

        if (n_per_epoch > 0 or n_per_epoch == -1) and not score_filename:
            adata = train_adata.concatenate(valid_adata)

            train_celltypes_encoded, _ = label_encoder(train_adata, le=None, condition_key=cell_type_key)
            valid_celltypes_encoded, _ = label_encoder(valid_adata, le=None, condition_key=cell_type_key)
            celltype_labels = np.concatenate([train_celltypes_encoded, valid_celltypes_encoded], axis=0)

            callbacks.append(ScoreCallback(score_filename, adata, condition_key, cell_type_key, self.cvae,
                                           n_per_epoch=n_per_epoch, n_batch_labels=len(self.n_conditions),
                                           n_celltype_labels=len(np.unique(celltype_labels))))

        if early_stop_limit > 0:
            callbacks.append(EarlyStopping(patience=early_stop_limit, monitor='val_loss'))

        if lr_reducer > 0:
            callbacks.append(ReduceLROnPlateau(monitor='val_loss', patience=lr_reducer))

        train_dataset, self.condition_encoder = make_dataset(train_adata, condition_key, self.condition_encoder,
                                                             batch_size, n_epochs, True,
                                                             self.loss_fn, self.n_conditions, self.size_factor_key)
        valid_dataset, _ = make_dataset(valid_adata, condition_key, self.condition_encoder, valid_adata.shape[0],
                                        n_epochs, False,
                                        self.loss_fn, self.n_conditions, self.size_factor_key)

        self.fit(train_dataset,
                 validation_data=valid_dataset,
                 epochs=n_epochs,
                 batch_size=batch_size,
                 verbose=fit_verbose,
                 callbacks=callbacks,
                 steps_per_epoch=steps_per_epoch,
                 validation_steps=1,
                 )
        if save:
            self.update_kwargs()
            self.save(make_dir=True)



import os
import warnings

from scarches.models import Adaptor

warnings.filterwarnings('ignore')

from scarches.zenodo.file import download_file
from scarches.zenodo.zip import unzip_model_directory

import numpy as np
from typing import Union

from keras import backend as K

from . import metrics
from . import models
from . import plotting as pl
from . import utils
from . import data
from . import annotation as ann
from . import datasets

__author__ = ', '.join([
    'Mohsen Naghipourfar',
    'Mohammad Lotfollahi',
    'Matin Khajavi',
])

__email__ = ', '.join([
    'naghipourfar@ce.sharif.edu',
    'mohammad.lotfollahi@helmholtz-muenchen.de',
])


def operate(network: Union[models.scArches, models.CVAE, models.scArchesNB, models.scArchesZINB],
            new_task_name: str,
            new_conditions: Union[list, str],
            adaptors: list = [],
            init: str = 'Xavier',
            version='scArches',
            remove_dropout: bool = True,
            print_summary: bool = False,
            new_training_kwargs: dict = {},
            new_network_kwargs: dict = {},
            ) -> Union[models.scArches, models.CVAE, models.scArchesNB, models.scArchesZINB]:
    """Performs architecture surgery on the pre-trained `network`.

        Parameters
        ----------
        network: :class:`~scarches.models.*`
            scArches Network object.
        new_task_name: str
            Name of the query task you want fine-tune the model on.
        new_conditions: list or str
            list of new conditions (studies or domains) in the query dataset. If only one condition exists in
            query dataset, Can pass a string to this argument.
        adaptors: list of `scarches.models.Adaptor` objects
            list of pre-trained adaptors to be attached to the reference network.
        init: str
            Method used for initializing new weights of the new model.
        version: str
            Version of scArches you want to use after performing surgery. Can be one of `scArches`, `scArches v1` and `scArches v2`.
        remove_dropout: bool
            Whether to remove the dropout layers.
        print_summary: bool
            Whether to print the summary of the new model.
        new_training_kwargs: dict
            Dictionary containing new training hyper-parameters used to compile and train model with.
        new_network_kwargs: dict
            Dictionary containing new network configuration used to create the new model.

        Returns
        -------
        new_network: :class:`~scarches.models.*`
            object of the new model.
    """
    if len(adaptors) > 0:
        network = attach_adaptors(network,
                                  adaptors,
                                  new_task_name=f"{network.task_name}_with_adaptors",
                                  remove_dropout=remove_dropout,
                                  print_summary=False,
                                  )

    version = version.lower()
    if version == 'scarches':
        freeze = True
        freeze_expression_input = True
    elif version == 'scarches v1':
        freeze = True
        freeze_expression_input = False
    elif version == 'scarches v2':
        freeze = False
        freeze_expression_input = False
    else:
        raise Exception("Invalid scArches version. Must be one of \'scArches\', \'scArches v1\', or \'scArches v2\'.")

    if isinstance(new_conditions, str):
        new_conditions = [new_conditions]

    n_new_conditions = len(new_conditions)
    network.update_kwargs()

    network_kwargs = network.network_kwargs
    training_kwargs = network.training_kwargs

    network_kwargs['conditions'] += sorted(new_conditions)

    if network_kwargs.get("mmd_computation_method", None):
        network_kwargs['mmd_computation_method'] = "general"

    network_kwargs['freeze_expression_input'] = freeze_expression_input

    if remove_dropout:
        network_kwargs['dropout_rate'] = 0.0

    for key in new_training_kwargs.keys():
        training_kwargs[key] = new_training_kwargs[key]

    for key in new_network_kwargs.keys():
        network_kwargs[key] = new_network_kwargs[key]

    # Instantiate new model with old parameters except `n_conditions`
    new_network = type(network)(task_name=new_task_name, **network_kwargs, **training_kwargs,
                                print_summary=False)

    # Get Previous Model's weights

    prev_biases_encoder = None
    for w in network.get_layer("encoder").get_layer("first_layer").weights:
        if "condition_kernel" in w.name:
            prev_condition_weights_encoder = K.get_value(w)
        elif "expression_kernel" in w.name:
            prev_input_weights_encoder = K.get_value(w)
        else:
            prev_biases_encoder = K.get_value(w)

    prev_biases_decoder = None
    for w in network.get_layer("decoder").get_layer("first_layer").weights:
        if "condition_kernel" in w.name:
            prev_condition_weights_decoder = K.get_value(w)
        elif "expression_kernel" in w.name:
            prev_latent_weights_decoder = K.get_value(w)
        else:
            prev_biases_decoder = K.get_value(w)

    # Modify the weights of 1st encoder & decoder layers
    if init == 'ones':
        to_be_added_weights_encoder_condition = np.ones(
            shape=(n_new_conditions, prev_condition_weights_encoder.shape[1]))
        to_be_added_weights_decoder_condition = np.ones(
            shape=(n_new_conditions, prev_condition_weights_decoder.shape[1]))
    elif init == "zeros":
        to_be_added_weights_encoder_condition = np.zeros(
            shape=(n_new_conditions, prev_condition_weights_encoder.shape[1]))
        to_be_added_weights_decoder_condition = np.zeros(
            shape=(n_new_conditions, prev_condition_weights_decoder.shape[1]))
    elif init == "Xavier":
        to_be_added_weights_encoder_condition = np.random.randn(n_new_conditions,
                                                                prev_condition_weights_encoder.shape[1]) * np.sqrt(
            2 / (prev_condition_weights_encoder.shape[0] + 1 + prev_condition_weights_encoder.shape[1]))
        to_be_added_weights_decoder_condition = np.random.randn(n_new_conditions,
                                                                prev_condition_weights_decoder.shape[1]) * np.sqrt(
            2 / (prev_condition_weights_decoder.shape[0] + 1 + prev_condition_weights_decoder.shape[1]))
    else:
        raise Exception("Invalid initialization for new weights")

    new_condition_weights_encoder = np.concatenate(
        [prev_condition_weights_encoder, to_be_added_weights_encoder_condition], axis=0)
    new_condition_weights_decoder = np.concatenate(
        [prev_condition_weights_decoder, to_be_added_weights_decoder_condition], axis=0)

    # Set new model's weights
    for w in new_network.get_layer("encoder").get_layer("first_layer").weights:
        if "condition_kernel" in w.name:
            K.set_value(w, new_condition_weights_encoder)
        elif "expression_kernel" in w.name:
            K.set_value(w, prev_input_weights_encoder)
        else:
            K.set_value(w, prev_biases_encoder)

    for w in new_network.get_layer("decoder").get_layer("first_layer").weights:
        if "condition_kernel" in w.name:
            K.set_value(w, new_condition_weights_decoder)
        elif "expression_kernel" in w.name:
            K.set_value(w, prev_latent_weights_decoder)
        else:
            K.set_value(w, prev_biases_decoder)

    # set weights of other parts of model
    for idx, encoder_layer in enumerate(new_network.encoder.layers):
        if encoder_layer.name != 'first_layer' and encoder_layer.get_weights() != []:
            encoder_layer.set_weights(network.encoder.layers[idx].get_weights())

    for idx, decoder_layer in enumerate(new_network.decoder.layers):
        if decoder_layer.name != 'first_layer' and decoder_layer.get_weights():
            decoder_layer.set_weights(network.decoder.layers[idx].get_weights())

    # Freeze old parts of cloned network
    if freeze:
        for encoder_layer in new_network.encoder.layers:
            if encoder_layer.name != 'first_layer':
                encoder_layer.trainable = False

        for decoder_layer in new_network.decoder.layers:
            if decoder_layer.name != 'first_layer':
                decoder_layer.trainable = False

        new_network.compile_models()

    # Print summary of new network
    if print_summary:
        new_network.get_summary_of_networks()

    # Add new condition to new network condition encoder
    new_network.condition_encoder = network.condition_encoder
    for idx, new_condition in enumerate(new_conditions):
        new_network.condition_encoder[new_condition] = network.n_conditions + idx

    return new_network


def create_scArches_from_pretrained_task(path_or_link: str,
                                         prev_task_name: str,
                                         model_path: str,
                                         new_task: str,
                                         target_conditions: list,
                                         version: str = 'scArches',
                                         **kwargs,
                                         ):
    """Performs architecture surgery on the pre-trained `network`.

        Parameters
        ----------
        path_or_link: str
            Path to the zip file of pre-trained model or direct downloadable link.
        prev_task_name: str
            Name of the previous task name the model pre-trained on.
        model_path: str
            Path to save the new model and extract the pre-trained model.
        new_task: str
            Name of the query task you want to train the model on.
        target_conditions: list
            list of new conditions (studies or domains) in the query dataset.
        version: str
            Version of scArches you want to use after performing surgery. Can be one of `scArches`, `scArches v1` and `scArches v2`.

        Returns
        -------
        new_network: :class:`~scarches.models.*`
            object of the new model.
    """
    if not os.path.isdir(path_or_link):
        downloaded_path = os.path.join(model_path, prev_task_name + ".zip")
        downloaded_path = __download_pretrained_scArches(path_or_link, save_path=downloaded_path, make_dir=True)
    else:
        downloaded_path = path_or_link

    if os.path.exists(downloaded_path) and downloaded_path.endswith(".zip"):
        extract_dir = os.path.join(model_path, f"{prev_task_name}/")
        unzip_model_directory(downloaded_path, extract_dir=extract_dir)
    elif not os.path.isdir(downloaded_path):
        raise ValueError("`model_path` should be either path to downloaded zip file or scArches pre-trained directory")

    downloaded_files = os.listdir(extract_dir)
    for file in downloaded_files:
        if file.startswith("scArches") and file.endswith(".json"):
            config_filename = file

    config_path = os.path.join(extract_dir, config_filename)
    pre_trained_scArches = models.scArches.from_config(config_path, new_params=kwargs, construct=True, compile=True)

    pre_trained_scArches.model_path = os.path.join(model_path, f"{prev_task_name}/")
    pre_trained_scArches.task_name = prev_task_name

    pre_trained_scArches.restore_model_weights(compile=True)

    scArches = operate(pre_trained_scArches,
                       new_conditions=target_conditions,
                       new_task_name=new_task,
                       init='Xavier',
                       version=version,
                       remove_dropout=False,
                       print_summary=False,
                       )

    scArches.model_path = os.path.join(model_path, f"{new_task}/")

    return scArches


def __download_pretrained_scArches(download_link: str,
                                   save_path: str = './',
                                   make_dir=False):
    if download_link != '':
        file_path, response = download_file(download_link, save_path, make_dir)
        return file_path
    else:
        raise Exception("Download link does not exist for the specified task")


def attach_adaptors(reference_model: Union[models.scArches, models.CVAE, models.scArchesNB, models.scArchesZINB],
                    adaptors: list,
                    new_task_name: str,
                    remove_dropout: bool = False,
                    print_summary: bool = False,
                    new_training_kwargs: dict = {},
                    new_network_kwargs: dict = {},
                    ) -> Union[models.scArches, models.CVAE, models.scArchesNB, models.scArchesZINB]:
    """Performs architecture surgery on the pre-trained `network`.

        Parameters
        ----------
        network: :class:`~scarches.models.*`
            scArches Network object.
        new_task_name: str
            Name of the query task you want fine-tune the model on.
        adaptors: list of Adaptor objects
            list of new conditions (studies or domains) in the query dataset. If only one condition exists in
            query dataset, Can pass a string to this argument.
        remove_dropout: bool
            Whether to remove the dropout layers.
        print_summary: bool
            Whether to print the summary of the new model.
        new_training_kwargs: dict
            Dictionary containing new training hyper-parameters used to compile and train model with.
        new_network_kwargs: dict
            Dictionary containing new network configuration used to create the new model.

        Returns
        -------
        new_network: :class:`~scarches.models.*`
            object of the new model.
    """
    new_conditions = [adaptor.study for adaptor in adaptors]
    if isinstance(new_conditions, str):
        new_conditions = [new_conditions]

    n_new_conditions = len(new_conditions)
    reference_model.update_kwargs()

    network_kwargs = reference_model.network_kwargs
    training_kwargs = reference_model.training_kwargs

    network_kwargs['conditions'] += sorted(new_conditions)

    if network_kwargs.get("mmd_computation_method", None):
        network_kwargs['mmd_computation_method'] = "general"

    network_kwargs['freeze_expression_input'] = True

    training_kwargs['model_path'] = reference_model.model_path.split(reference_model.task_name)[0]
    training_kwargs['gamma'] = 0.0

    if remove_dropout:
        network_kwargs['dropout_rate'] = 0.0

    for key in new_training_kwargs.keys():
        training_kwargs[key] = new_training_kwargs[key]

    for key in new_network_kwargs.keys():
        network_kwargs[key] = new_network_kwargs[key]

    # Instantiate new model with old parameters except `n_conditions`
    new_network = type(reference_model)(task_name=new_task_name, **network_kwargs, **training_kwargs,
                                        print_summary=False)

    # Get Previous Model's weights
    used_bias_encoder = reference_model.get_layer("encoder").get_layer("first_layer").use_bias
    used_bias_decoder = reference_model.get_layer("decoder").get_layer("first_layer").use_bias

    prev_weights = {}
    for w in reference_model.get_layer("encoder").get_layer("first_layer").weights:
        if "condition_kernel" in w.name:
            prev_weights['c'] = K.get_value(w)
        elif "expression_kernel" in w.name:
            prev_weights['i'] = K.get_value(w)
        else:
            prev_weights['b'] = K.get_value(w)

    if used_bias_encoder:
        prev_input_weights_encoder, prev_condition_weights_encoder, prev_biases_encoder = \
            prev_weights['i'], prev_weights['c'], prev_weights['b']

    else:
        prev_input_weights_encoder, prev_condition_weights_encoder, prev_biases_encoder = \
            prev_weights['i'], prev_weights['c'], None

    prev_weights = {}
    for w in reference_model.get_layer("decoder").get_layer("first_layer").weights:
        if "condition_kernel" in w.name:
            prev_weights['c'] = K.get_value(w)
        elif "expression_kernel" in w.name:
            prev_weights['i'] = K.get_value(w)
        else:
            prev_weights['b'] = K.get_value(w)

    if used_bias_decoder:
        prev_latent_weights_decoder, prev_condition_weights_decoder, prev_biases_decoder = \
            prev_weights['i'], prev_weights['c'], prev_weights['b']
    else:
        prev_latent_weights_decoder, prev_condition_weights_decoder, prev_biases_decoder = \
            prev_weights['i'], prev_weights['c'], None

    # Modify the weights of 1st encoder & decoder layers
    to_be_added_weights_encoder_condition = np.vstack([adaptor.encoder_weights for adaptor in adaptors])
    to_be_added_weights_decoder_condition = np.vstack([adaptor.decoder_weights for adaptor in adaptors])

    new_condition_weights_encoder = np.concatenate(
        [prev_condition_weights_encoder, to_be_added_weights_encoder_condition], axis=0)
    new_condition_weights_decoder = np.concatenate(
        [prev_condition_weights_decoder, to_be_added_weights_decoder_condition], axis=0)

    # Set new model's weights
    for w in new_network.get_layer("encoder").get_layer("first_layer").weights:
        if "condition_kernel" in w.name:
            K.set_value(w, new_condition_weights_encoder)
        elif "expression_kernel" in w.name:
            K.set_value(w, prev_input_weights_encoder)
        else:
            K.set_value(w, prev_biases_encoder)

    for w in new_network.get_layer("decoder").get_layer("first_layer").weights:
        if "condition_kernel" in w.name:
            K.set_value(w, new_condition_weights_decoder)
        elif "expression_kernel" in w.name:
            K.set_value(w, prev_latent_weights_decoder)
        else:
            K.set_value(w, prev_biases_decoder)

    # set weights of other parts of model
    for idx, encoder_layer in enumerate(new_network.encoder.layers):
        if encoder_layer.name != 'first_layer' and encoder_layer.get_weights() != []:
            encoder_layer.set_weights(reference_model.encoder.layers[idx].get_weights())

    for idx, decoder_layer in enumerate(new_network.decoder.layers):
        if decoder_layer.name != 'first_layer' and decoder_layer.get_weights():
            decoder_layer.set_weights(reference_model.decoder.layers[idx].get_weights())

    # Freeze old parts of cloned network
    for encoder_layer in new_network.encoder.layers:
        if encoder_layer.name != 'first_layer':
            encoder_layer.trainable = False

    for decoder_layer in new_network.decoder.layers:
        if decoder_layer.name != 'first_layer':
            decoder_layer.trainable = False

    new_network.compile_models()

    # Print summary of new network
    if print_summary:
        new_network.get_summary_of_networks()

    # Add new condition to new network condition encoder
    new_network.condition_encoder = reference_model.condition_encoder
    for idx, new_condition in enumerate(new_conditions):
        new_network.condition_encoder[new_condition] = reference_model.n_conditions + idx

    return new_network


def extract_adaptors(network: Union[models.scArches, models.scArchesNB, models.scArchesZINB],
                     conditions: Union[list, str]):
    """
    Extracts new adaptors from the trained network.

    Parameters
    ----------
    network: :class:`~scarches.models.*`
        Trained network object. Has to be object of one of `scArches`, `scArchesNB`, and `scArchesZINB` class.

    conditions: list
        List of conditions (studies).

    Returns
    -------
    adaptors: list
        List of query adaptors.

    """
    if isinstance(conditions, str):
        conditions = [conditions]

    condition_indices = [network.condition_encoder[new_condition] for new_condition in conditions]

    for w in network.get_layer("encoder").get_layer("first_layer").weights:
        if "condition_kernel" in w.name:
            condition_weights_encoder = K.get_value(w)

    for w in network.get_layer("decoder").get_layer("first_layer").weights:
        if "condition_kernel" in w.name:
            condition_weights_decoder = K.get_value(w)

    adaptors = []
    for i in range(len(conditions)):
        adaptor = Adaptor(study=conditions[i],
                          encoder_weights=condition_weights_encoder[condition_indices[i]],
                          decoder_weights=condition_weights_decoder[condition_indices[i]]
                          )

        adaptors.append(adaptor)

    return adaptors


def download_adaptor(download_link: str,
                     path_to_save: str,
                     make_dir: bool = False):
    """Downloads an adaptor via given `download_link`.

    Parameters
    ----------
    download_link: str
        Direct downloadable link.
    path_to_save: str
        Path to save the adaptor.
    make_dir: bool
        Whether to make directory if it does not exist.

    Returns
    -------
    adaptor: :class:`~scarches.models.Adaptor`
        object of the new Adaptor.
        """
    import pickle
    if download_link != "":
        file_path, response = download_file(download_link, path_to_save, make_dir=make_dir)
        adaptor = pickle.load(open(file_path, 'rb'))
        return adaptor



from pathlib import Path

from setuptools import setup, find_packages

long_description = Path('README.rst').read_text('utf-8')

try:
    from scarches import __author__, __email__
except ImportError:  # Deps not yet installed
    __author__ = __email__ = ''

setup(name='scArches',
      version='0.2.2',
      description='Transfer learning with Architecture Surgery on Single-cell data',
      long_description=long_description,
      long_description_content_type="text/markdown",
      url='https://github.com/theislab/scarches',
      author=__author__,
      author_email=__email__,
      license='MIT',
      packages=find_packages(),
      zip_safe=False,
      install_requires=[
          l.strip() for l in Path('requirements.txt').read_text('utf-8').splitlines()
      ],
      classifiers=[
          "Programming Language :: Python :: 3.6",
          "Programming Language :: Python :: 3.7",
          'Environment :: Console',
          'Framework :: Jupyter',
          'Intended Audience :: Developers',
          'Intended Audience :: Science/Research',
          "License :: OSI Approved :: MIT License",
          'Natural Language :: English',
          'Operating System :: MacOS :: MacOS X',
          'Operating System :: Microsoft :: Windows',
          'Operating System :: POSIX :: Linux',
      ],
      doc=[
          'sphinx',
          'sphinx_rtd_theme',
          'sphinx_autodoc_typehints',
          'typing_extensions; python_version < "3.8"',
      ],
      )



import os
import csv
from typing import TypeVar

import numpy as np
import pandas as pd
import time
import anndata
from tensorflow.keras.callbacks import Callback
from tensorflow.keras.models import Model
from tensorflow.keras.utils import to_categorical
from scipy.stats import entropy, itemfreq
from sklearn.cluster import KMeans
from sklearn.metrics import silhouette_score, adjusted_rand_score, normalized_mutual_info_score
from sklearn.neighbors import NearestNeighbors
from scarches.utils import remove_sparsity, label_encoder

list_or_str = TypeVar('list_or_str', list, str)


class ScoreCallback(Callback):
    def __init__(self,
                 filename: str,
                 adata: anndata.AnnData,
                 batch_key: str,
                 cell_type_key: str,
                 encoder_model: Model,
                 n_per_epoch: int = 5,
                 n_batch_labels: int = 0,
                 n_celltype_labels: int = 0,
                 clustering_scores: list_or_str = 'all'
                 ):
        super(ScoreCallback, self).__init__()
        self.adata = remove_sparsity(adata)

        self.batch_labels, _ = label_encoder(adata, le=None, condition_key=batch_key)
        self.batch_labels = np.reshape(self.batch_labels, (-1,))
        self.batch_labels_onehot = to_categorical(self.batch_labels, num_classes=n_batch_labels)

        self.celltype_labels, _ = label_encoder(adata, le=None, condition_key=cell_type_key)
        self.celltype_labels = np.reshape(self.celltype_labels, (-1,))
        self.celltype_labels_onehot = to_categorical(self.celltype_labels, num_classes=n_celltype_labels)

        self.filename = filename
        self.encoder_model = encoder_model
        self.n_per_epoch = n_per_epoch

        self.n_batch_labels = n_batch_labels
        self.n_celltype_labels = n_celltype_labels

        self.clustering_scores = clustering_scores
        self.score_computers = {"asw": self.asw,
                                "ari": self.ari,
                                "nmi": self.nmi,
                                "ebm": self.entropy_of_batch_mixing,
                                "knn": self.knn_purity}

        self.kmeans_batch = KMeans(self.n_batch_labels, n_init=200)
        self.kmeans_celltype = KMeans(self.n_celltype_labels, n_init=200)

    def on_train_begin(self, logs=None):
        self.scores = []
        self.epochs = []
        self.times = []

        if self.clustering_scores == 'all':
            clustering_scores_columns = ['epoch', 'time', 'ASW', 'ARI', 'NMI', 'EBM', 'KNN']

        os.makedirs(os.path.dirname(self.filename), exist_ok=True)
        if not os.path.exists(self.filename):
            with open(self.filename, 'a') as f:
                writer = csv.writer(f)
                writer.writerow(clustering_scores_columns)

    def on_train_end(self, logs=None):
        latent_X = self.encoder_model.predict([self.adata.X, self.batch_labels_onehot, self.batch_labels_onehot])[1]
        self.epochs.append(-1)
        last_time_record = self.times[-1] if len(self.times) > 0 else 0.0
        print(f"Last Epoch: ", end="\t")
        asw, asw_time = self.asw(latent_X)
        ari, ari_time = self.ari(latent_X)
        nmi, nmi_time = self.nmi(latent_X)
        ebm, ebm_time = self.entropy_of_batch_mixing(latent_X, n_neighbors=15, n_pools=1)
        knn, knn_time = self.knn_purity(latent_X, n_neighbors=15)
        self.scores.append([asw, ari, nmi, ebm, knn])
        print(
            f"ASW: {asw:.4f} - ARI: {ari:.4f} - NMI: {nmi:.4f} - EBM: {ebm:.4f} - KNN_Purity: {knn: .4f} - {latent_X.shape}")
        computation_times = asw_time + ari_time + nmi_time + ebm_time + knn_time

        self.times.append(time.time() - self.epoch_time_start + last_time_record - computation_times)

        scores_df = pd.DataFrame({"epoch": self.epochs, "time": self.times})

        self.scores_np = np.array(self.scores)

        if self.clustering_scores == 'all':
            self.clustering_scores = ['ASW', 'ARI', 'NMI', 'EBM', 'KNN']
        for i, clustering_score in enumerate(self.clustering_scores):
            computed_scores = self.scores_np[:, i]
            scores_df[clustering_score] = computed_scores
        
        with open(self.filename, 'a') as f:
            writer = csv.writer(f)
            for i in range(scores_df.shape[0]):
                writer.writerow(scores_df.values[i, :])
        # scores_df.to_csv(self.filename, index=False)

    def on_epoch_begin(self, epoch, logs=None):
        self.epoch_time_start = time.time()

    def on_epoch_end(self, epoch, logs=None):
        if epoch % self.n_per_epoch == 0 and self.n_per_epoch > 0:
            latent_X = self.encoder_model.predict([self.adata.X, self.batch_labels_onehot, self.batch_labels_onehot])[1]

            self.epochs.append(epoch)
            last_time_record = self.times[-1] if len(self.times) > 0 else 0.0
            print(f"Epoch {epoch}: ", end="\t")
            if self.clustering_scores == 'all':
                asw, asw_time = self.asw(latent_X)
                ari, ari_time = self.ari(latent_X)
                nmi, nmi_time = self.nmi(latent_X)
                ebm, ebm_time = self.entropy_of_batch_mixing(latent_X, n_neighbors=15)
                knn, knn_time = self.knn_purity(latent_X, n_neighbors=15)
                self.scores.append([asw, ari, nmi, ebm, knn])
                print(
                    f"ASW: {asw:.4f} - ARI: {ari:.4f} - NMI: {nmi:.4f} - EBM: {ebm:.4f} - KNN_Purity: {knn: .4f} - {latent_X.shape}")
                computation_times = asw_time + ari_time + nmi_time + ebm_time + knn_time
            else:
                scores = []
                computation_times = 0
                for clustering_score in self.clustering_scores:
                    score, computation_time = self.score_computers[clustering_score](latent_X)
                    scores.append(score)
                    computation_times += computation_time
                    print(f"{clustering_score}: {scores[-1]:.4f}", end=" - ")
                print()
                self.scores.append(scores)

            self.times.append(time.time() - self.epoch_time_start + last_time_record - computation_times)

    def asw(self, latent):
        start_time = time.time()
        score = silhouette_score(latent, self.batch_labels)
        end_time = time.time()
        return score, end_time - start_time

    def ari(self, latent):
        start_time = time.time()
        labels_pred = self.kmeans_celltype.fit_predict(latent)
        score = adjusted_rand_score(self.celltype_labels, labels_pred)
        end_time = time.time()
        return score, end_time - start_time

    def nmi(self, latent):
        start_time = time.time()
        labels_pred = self.kmeans_celltype.fit_predict(latent)
        score = normalized_mutual_info_score(self.celltype_labels, labels_pred)
        end_time = time.time()
        return score, end_time - start_time

    def entropy_of_batch_mixing(self, latent,
                                n_neighbors=50, n_pools=50, n_samples_per_pool=100):
        start_time = time.time()

        def entropy_from_indices(indices):
            return entropy(np.array(itemfreq(indices)[:, 1].astype(np.int32)))

        n_samples = latent.shape[0]

        neighbors = NearestNeighbors(n_neighbors=n_neighbors + 1).fit(latent)
        indices = neighbors.kneighbors(latent, return_distance=False)[:, 1:]
        batch_indices = np.vectorize(lambda i: self.batch_labels[i])(indices)

        entropies = np.apply_along_axis(entropy_from_indices, axis=1, arr=batch_indices)

        # average n_pools entropy results where each result is an average of n_samples_per_pool random samples.
        if n_pools == 1:
            score = np.mean(entropies)
        else:
            score = np.mean([
                np.mean(entropies[np.random.choice(len(entropies), size=n_samples_per_pool)])
                for _ in range(n_pools)
            ])
        end_time = time.time()
        return score, end_time - start_time

    def knn_purity(self, latent, n_neighbors=30):
        start_time = time.time()

        nbrs = NearestNeighbors(n_neighbors=n_neighbors + 1).fit(latent)
        indices = nbrs.kneighbors(latent, return_distance=False)[:, 1:]
        neighbors_labels = np.vectorize(lambda i: self.celltype_labels[i])(indices)

        # pre cell purity scores
        scores = ((neighbors_labels - self.celltype_labels.reshape(-1, 1)) == 0).mean(axis=1)
        res = [
            np.mean(scores[self.celltype_labels == i]) for i in np.unique(self.celltype_labels)
        ]  # per cell-type purity

        end_time = time.time()

        return np.mean(res), end_time - start_time
